-----------------------------------------------------------------------------------------
{-| Module      :  Database.HSQL.Oracle
    Copyright   :  (c) Krasimir Angelov 2003
    License     :  BSD-style

    Maintainer  :  kr.angelov@gmail.com
    Stability   :  provisional
    Portability :  portable

    The module provides interface to Oracle
-}
-----------------------------------------------------------------------------------------

module Database.HSQL.Oracle(connect, module HSQL) where

import qualified Database.HSQL as HSQL
import Database.HSQL.Types
import Data.Word (Word8, Word16, Word32, )
import Foreign.Ptr
import Foreign.ForeignPtr
import Foreign.Storable
import Foreign.Marshal
import Foreign.C
         (CInt, CShort, peekCStringLen, withCStringLen,
          CString, peekCString, newCString, )
import qualified Foreign.Concurrent as FC
import Control.Concurrent.MVar (newMVar, )
import Control.Exception(catchDyn, throwDyn, bracket, )
import Data.List(foldl')

#include <HsOCI.h>

type OCIHandle a = Ptr a

newtype RecEnv      = RecEnv      (); type OCIEnv      = OCIHandle RecEnv
newtype RecError    = RecError    (); type OCIError    = OCIHandle RecError
newtype RecSvcCtx   = RecSvcCtx   (); type OCISvcCtx   = OCIHandle RecSvcCtx
newtype RecStmt     = RecStmt     (); type OCIStmt     = OCIHandle RecStmt
newtype RecParam    = RecParam    (); type OCIParam    = OCIHandle RecParam
newtype RecDefine   = RecDefine   (); type OCIDefine   = OCIHandle RecDefine
newtype RecBind     = RecBind     (); type OCIBind     = OCIHandle RecBind
newtype RecDescribe = RecDescribe (); type OCIDescribe = OCIHandle RecDescribe

type OCIEnvRef = ForeignPtr RecEnv

newtype OCIResult = OCIResult CInt
   deriving (Eq, Show)


newtype OCIType t = OCIType CInt

class OCIRec t where typeIdent :: OCIType t

instance OCIRec RecStmt     where typeIdent = OCIType (#const OCI_HTYPE_STMT)
instance OCIRec RecParam    where typeIdent = OCIType (#const OCI_DTYPE_PARAM)
instance OCIRec RecDescribe where typeIdent = OCIType (#const OCI_HTYPE_DESCRIBE)
instance OCIRec RecError    where typeIdent = OCIType (#const OCI_HTYPE_ERROR)


foreign import ccall "OCIEnvCreate" ociEnvCreate :: Ptr OCIEnv -> CInt -> Ptr a -> FunPtr a -> FunPtr a -> FunPtr a -> CInt -> Ptr (Ptr a) -> IO OCIResult
foreign import ccall "OCITerminate" ociTerminate :: CInt -> IO OCIResult
foreign import ccall "OCIHandleAlloc" ociHandleAlloc :: OCIHandle a -> Ptr (OCIHandle b) -> OCIType b -> CInt -> Ptr a -> IO OCIResult
foreign import ccall "OCIHandleFree" ociHandleFree :: OCIHandle b -> OCIType b -> IO OCIResult
foreign import ccall "OCIErrorGet" ociErrorGet :: OCIHandle a -> CInt -> CString -> Ptr CInt -> CString -> CInt -> CInt -> IO OCIResult

foreign import ccall "OCILogon" ociLogon :: OCIEnv -> OCIError -> Ptr OCISvcCtx -> CString -> CInt -> CString -> CInt -> CString -> CInt -> IO OCIResult
foreign import ccall "OCILogoff" ociLogoff :: OCISvcCtx -> OCIError -> IO OCIResult

foreign import ccall "OCIStmtPrepare" ociStmtPrepare :: OCIStmt -> OCIError -> CString -> CInt -> CInt -> CInt -> IO OCIResult
foreign import ccall "OCIStmtExecute" ociStmtExecute :: OCISvcCtx -> OCIStmt -> OCIError -> CInt -> CInt -> OCIHandle () -> OCIHandle () -> CInt -> IO OCIResult
foreign import ccall "OCIStmtFetch2" ociStmtFetch2 :: OCIStmt -> OCIError -> CInt -> CInt -> CInt -> CInt -> IO OCIResult
foreign import ccall "OCIDefineByPos" ociDefineByPos :: OCIStmt -> Ptr OCIDefine -> OCIError -> CInt -> Ptr a -> CInt -> CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> CInt -> IO OCIResult
foreign import ccall "OCIBindByPos"   ociBindByPos   :: OCIStmt -> Ptr OCIBind   -> OCIError -> CInt -> Ptr a -> CInt -> CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> CInt -> Ptr CInt -> CInt -> IO OCIResult

foreign import ccall "OCIParamGet" ociParamGet :: OCIHandle a -> OCIType a -> OCIError -> Ptr OCIParam -> CInt -> IO OCIResult
foreign import ccall "OCIAttrGet" ociAttrGet :: OCIHandle b -> OCIType b -> Ptr a -> Ptr CInt -> CInt -> OCIError -> IO OCIResult
foreign import ccall "OCIAttrSet" ociAttrSet :: OCIHandle b -> OCIType b -> Ptr a -> CInt -> CInt -> OCIError -> IO OCIResult
foreign import ccall "OCIDescribeAny" ociDescribeAny :: OCISvcCtx -> OCIError -> Ptr () -> CInt -> Word8 -> Word8 -> Word8 -> OCIDescribe -> IO OCIResult

foreign import ccall "OCIDescriptorFree" ociDescriptorFree :: OCIHandle a -> CInt -> IO OCIResult

foreign import ccall "OCITransStart" ociTransStart :: OCISvcCtx -> OCIError -> Word8 -> CInt -> IO OCIResult
foreign import ccall "OCITransCommit" ociTransCommit :: OCISvcCtx -> OCIError -> CInt -> IO OCIResult
foreign import ccall "OCITransRollback" ociTransRollback :: OCISvcCtx -> OCIError -> CInt -> IO OCIResult

foreign import ccall "strlen" strlen :: CString -> IO CInt

-----------------------------------------------------------------------------------------
-- keeper of OCIEnv
-----------------------------------------------------------------------------------------

getMyEnvironment :: IO OCIEnvRef
getMyEnvironment = do
  env <- create nullPtr $ \ pOCIEnv -> ociEnvCreate pOCIEnv (#const OCI_DEFAULT) nullPtr nullFunPtr nullFunPtr nullFunPtr 0 nullPtr
  FC.newForeignPtr env terminate
  where
    terminate = ociTerminate (#const OCI_DEFAULT) >>= handleSqlResult nullPtr

-----------------------------------------------------------------------------------------
-- error handling
-----------------------------------------------------------------------------------------

handleSqlResult :: OCIError -> OCIResult -> IO ()
handleSqlResult err res
  | res == OCIResult (#const OCI_SUCCESS) || res == OCIResult (#const OCI_NO_DATA) = return ()
  | res == OCIResult (#const OCI_SUCCESS_WITH_INFO) = do
#ifdef DEBUG
      e <- getSqlError
      putTraceMsg (show e)
#else
      return ()
#endif
  | res == OCIResult (#const OCI_INVALID_HANDLE)  = throwDyn SqlInvalidHandle
  | res == OCIResult (#const OCI_STILL_EXECUTING) = throwDyn SqlStillExecuting
  | res == OCIResult (#const OCI_NEED_DATA)       = throwDyn SqlNeedData
  | res == OCIResult (#const OCI_ERROR)           = getSqlError >>= throwDyn
  | otherwise = error (show res)
  where
    stringBufferLen = 1024
    
    getSqlError =
      alloca $ \pErrCode ->
      allocaBytes stringBufferLen $ \pErrMsg -> do
        OCIResult rc <- ociErrorGet err 1 nullPtr pErrCode pErrMsg (fromIntegral stringBufferLen) (#const OCI_HTYPE_ERROR)
        if rc < 0
          then return SqlNoData
          else do
            msg <- peekCString pErrMsg
            errCode <- peek pErrCode
            return (SqlError {seState="", seNativeError=fromIntegral errCode, seErrorMsg=msg})

-- | Makes a new connection to the Oracle service
connect :: String               -- ^ Service name
        -> String               -- ^ User identifier
        -> String               -- ^ Password
        -> IO Connection        -- ^ the returned value represents the new connection
connect service user pwd =
  let connector = connectPrivate service user pwd
  in  do connection <- connector
         return (connection{connTables =
                   catchDyn
                      -- first try to access COREDB_SYSTEM
                      (connTables connection)
                      -- if this does not exist, try USER_TABLES
                      (\exc -> case exc of SqlError _ _ _ -> tablesFromUserTables connector)
             })

-- | Connect, but the low-level implementation of connTables does not work in newer Oracle versions (e.g. 10g)
connectPrivate ::
           String               -- ^ Service name
        -> String               -- ^ User identifier
        -> String               -- ^ Password
        -> IO Connection        -- ^ the returned value represents the new connection
connectPrivate service0 user0 pwd0 =
  getMyEnvironment >>= \myEnvironment ->
  withForeignPtr myEnvironment $ \env ->
  withCStringLen user0    $ \(user, user_len) ->
  withCStringLen pwd0     $ \(pwd, pwd_len) ->
  withCStringLen service0 $ \(service, service_len) ->
  do
    err <- create nullPtr (\pError -> ociHandleAlloc env pError typeIdent 0 nullPtr)
    svcCtx <- create err (\pSvcCtx -> ociLogon env err pSvcCtx user (fromIntegral user_len) pwd (fromIntegral pwd_len) service (fromIntegral service_len))
    refFalse <- newMVar False
    let connection = (Connection
    			{ connDisconnect = disconnect svcCtx err
    			, connExecute    = execute myEnvironment svcCtx err
    			, connPrepare    = prepare connection myEnvironment svcCtx err
    			, connQuery      = query connection myEnvironment svcCtx err
    			, connTables     = tablesFromCoreDB env svcCtx err
    			, connDescribe   = describe env svcCtx err
    			, connBeginTransaction = beginTransaction myEnvironment svcCtx err
    			, connCommitTransaction = commitTransaction myEnvironment svcCtx err 
    			, connRollbackTransaction = rollbackTransaction myEnvironment svcCtx err
    			, connClosed     = refFalse
    			})
    return connection
  where
    disconnect svcCtx err = do
      ociLogoff svcCtx err >>= handleSqlResult err
      ociHandleFree err typeIdent >>= handleSqlResult err

    execute envRef svcCtx err query =
      withForeignPtr envRef $ \env ->
      withCStringLen query $ \(query,query_len) ->
      withHandle env err $ \stmt ->
      do
        ociStmtPrepare stmt err query (fromIntegral query_len) (#const OCI_NTV_SYNTAX) (#const OCI_DEFAULT) >>= handleSqlResult err
        ociStmtExecute svcCtx stmt err 1 0 nullPtr nullPtr (#const OCI_DEFAULT) >>= handleSqlResult err
        count <- create err (\pCount -> ociAttrGet stmt typeIdent pCount nullPtr (#const OCI_ATTR_ROW_COUNT) err)
        return $! fromIntegral (count :: (#type ub4))

    prepare connection envRef svcCtx err query = 
      withForeignPtr envRef $ \env ->
      withCStringLen query $ \(query,query_len) ->
      do
        stmt <- create err (\ pStmt -> ociHandleAlloc env pStmt typeIdent 0 nullPtr)
	ociStmtPrepare stmt err query (fromIntegral query_len) (#const OCI_NTV_SYNTAX) (#const OCI_DEFAULT) >>= handleSqlResult err
	fields <- allocaBytes (#const (sizeof(FIELD_DEF))) (getFieldDefs err stmt 1)
	let offsets_arr_size = length fields * sizeOf offsets_arr_size
	buffer <- mallocBytes (foldl' (\s t -> sizeOf (undefined :: CShort) + sqlType2Size t + s) offsets_arr_size fields)
	definePositions stmt err buffer 0 offsets_arr_size fields
	refFalse <- newMVar False
        return (Statement
                  { stmtConn   = connection
                  , stmtClose  = closeStatement stmt buffer err
                  , stmtExecute= executePrepared svcCtx stmt err
                  , stmtFetch  = fetch stmt err
                  , stmtGetCol = getColValue buffer
                  , stmtFields = fields
                  , stmtClosed = refFalse
                  , stmtPrefetch = setPrefetch stmt err
                  })

    query connection envRef svcCtx err query = 
      withForeignPtr envRef $ \env ->
      withCStringLen query $ \(query,query_len) ->
      do
        stmt <- create err (\ pStmt -> ociHandleAlloc env pStmt typeIdent 0 nullPtr)
	ociStmtPrepare stmt err query (fromIntegral query_len) (#const OCI_NTV_SYNTAX) (#const OCI_DEFAULT) >>= handleSqlResult err
	ociStmtExecute svcCtx stmt err 0 0 nullPtr nullPtr (#const OCI_DEFAULT) >>= handleSqlResult err
	fields <- allocaBytes (#const (sizeof(FIELD_DEF))) (getFieldDefs err stmt 1)
	let offsets_arr_size = length fields * sizeOf offsets_arr_size
	buffer <- mallocBytes (foldl' (\s t -> sizeOf (undefined :: CShort) + sqlType2Size t + s) offsets_arr_size fields)
	definePositions stmt err buffer 0 offsets_arr_size fields
	refFalse <- newMVar False
        return (Statement
                  { stmtConn   = connection
                  , stmtClose  = closeStatement stmt buffer err
                  , stmtExecute= throwDyn SqlUnsupportedOperation
                  , stmtFetch  = fetch stmt err
                  , stmtGetCol = getColValue buffer
                  , stmtFields = fields
                  , stmtClosed = refFalse
                  , stmtPrefetch = setPrefetch stmt err
                  })

    executePrepared svcCtx stmt err values =
      do
        bindings <- setParams stmt err 0 [] values
        res <- ociStmtExecute svcCtx stmt err 1 0 nullPtr nullPtr (#const OCI_DEFAULT)
        mapM_ free bindings
        handleSqlResult err res
        count <- create err $ \ pCount -> ociAttrGet stmt typeIdent pCount nullPtr (#const OCI_ATTR_ROW_COUNT) err
        return $! fromIntegral (count :: (#type ub4))

    setParams stmt err index bindings []                                 = return bindings
    setParams stmt err index bindings (SqlStringValue    value : values) =
      alloca $ \pBind -> do
        cstr <- newCString value
        ociBindByPos stmt pBind err (index+1) cstr (fromIntegral (length value+1)) (#const SQLT_STR) nullPtr nullPtr nullPtr (#const OCI_DEFAULT) nullPtr (#const OCI_DEFAULT) >>= handleSqlResult err
        setParams stmt err (index+1) (castPtr cstr:bindings) values
    setParams stmt err index bindings (SqlIntValue       value : values) =
      alloca $ \pBind -> do
        pValue <- malloc
        poke pValue value
        res <- ociBindByPos stmt pBind err (index+1) pValue (fromIntegral (sizeOf value)) (#const SQLT_INT) nullPtr nullPtr nullPtr (#const OCI_DEFAULT) nullPtr (#const OCI_DEFAULT)
        handleSqlResult err res
        setParams stmt err (index+1) (castPtr pValue:bindings) values
    setParams stmt err index bindings (SqlDoubleValue    value : values) =
      alloca $ \pBind -> do
        pValue <- malloc
        poke pValue value
        res <- ociBindByPos stmt pBind err (index+1) pValue (fromIntegral (sizeOf value)) (#const SQLT_FLT) nullPtr nullPtr nullPtr (#const OCI_DEFAULT) nullPtr (#const OCI_DEFAULT)
        handleSqlResult err res
        setParams stmt err (index+1) (castPtr pValue:bindings) values

    getFieldDefs err stmt counter buffer = do
      res <- ociParamGet stmt typeIdent err ((#ptr FIELD_DEF, par) buffer) counter
      if res == OCIResult (#const OCI_SUCCESS)
        then do field  <- getFieldDef err buffer
                fields <- getFieldDefs err stmt (counter+1) buffer
                return (field:fields)
        else return []

    sqlType2Size :: FieldDef -> Int
    sqlType2Size (_,tp,_) =
      case tp of
        SqlVarChar n   -> n+1
        SqlNumeric p s -> p+s+3   -- The value precision plus optional positions for '.', '-' and
                                  -- one position for the '\0' character at end of the string.
        SqlInteger     -> 16 -- 12 digits are enough (maxBound :: Int) has 10 digits.
                             -- in addition we may need one position for '-' and one
                             -- for the '\0' character at end of the string.
        SqlFloat       -> 100
        SqlDate        -> 100
	SqlTime        -> 100
	SqlTimeTZ      -> 100
	SqlTimeStamp   -> 100
	SqlText        -> 100
	SqlUnknown _   -> throwDyn SqlUnsupportedOperation

    definePositions stmt err buffer pos offset [] = return ()
    definePositions stmt err buffer pos offset (field:fields) =
      alloca $ \pDef -> do
        let size = sqlType2Size field
        poke (castPtr buffer `advancePtr` pos) (fromIntegral offset :: CInt)
        ociDefineByPos stmt
                       pDef
                       err
                       (fromIntegral pos+1)
                       (buffer `plusPtr` (offset+sizeOf (undefined :: CShort)))
                       (fromIntegral size)
                       (#const SQLT_STR)
                       (castPtr buffer `plusPtr` offset)
                       nullPtr
                       nullPtr
                       (#const OCI_DEFAULT)
        definePositions stmt err buffer (pos+1) (offset+size) fields

    mkSqlType :: (#type OCITypeCode) -> (#type ub2) -> (#type ub1) -> (#type ub1) -> SqlType
    mkSqlType (#const SQLT_CHR)       size _    _     = SqlVarChar (fromIntegral size)
    mkSqlType (#const SQLT_AFC)       size _    _     = SqlChar    (fromIntegral size)
    mkSqlType (#const SQLT_NUM)       _    prec scale = SqlNumeric (fromIntegral prec) (fromIntegral scale)
    mkSqlType (#const SQLT_INT)       _    _    _     = SqlInteger
    mkSqlType (#const SQLT_FLT)       _    _    _     = SqlFloat
    mkSqlType (#const SQLT_DATE)      _    _    _     = SqlDate
    mkSqlType (#const SQLT_TIME)      _    _    _     = SqlTime
    mkSqlType (#const SQLT_TIME_TZ)   _    _    _     = SqlTimeTZ
    mkSqlType (#const SQLT_TIMESTAMP) _    _    _     = SqlTimeStamp
    mkSqlType (#const SQLT_LNG)       _    _    _     = SqlText
    mkSqlType tp                      _    _    _     = SqlUnknown (fromIntegral tp)

    describe env svcCtx err tblName =
      withCStringLen tblName $ \(cstr,clen) ->
      withHandle env err $ \descr -> do
        ociDescribeAny svcCtx err (castPtr cstr) (fromIntegral clen) (#const OCI_OTYPE_NAME) (#const OCI_DEFAULT) (#const OCI_PTYPE_TABLE) descr >>= handleSqlResult err
        withAttr descr err (#const OCI_ATTR_PARAM) $ \param -> do
          numcols <- create err (\pNumcols -> ociAttrGet param typeIdent pNumcols nullPtr (#const OCI_ATTR_NUM_COLS) err)
          -- fieldDefs
          withAttr (param::OCIParam) err (#const OCI_ATTR_LIST_COLUMNS) $ \ coll ->
            allocaBytes (#const (sizeof(FIELD_DEF))) (getFieldDefs coll 1 (fromIntegral (numcols :: (#type ub2))))
      where
        getFieldDefs coll index numcols buffer
          | index <= numcols = do
               ociParamGet (coll::OCIParam) typeIdent err ((#ptr FIELD_DEF, par) buffer) index >>= handleSqlResult err
	       fieldDef  <- getFieldDef err buffer
               fieldDefs <- getFieldDefs coll (index+1) numcols buffer
               return (fieldDef:fieldDefs)
          | otherwise = return []

    getFieldDef err buffer = do
      par <- (#peek FIELD_DEF, par) buffer
      ociAttrGet (par::OCIParam) typeIdent ((#ptr FIELD_DEF, dtype) buffer) nullPtr (#const OCI_ATTR_DATA_TYPE) err >>= handleSqlResult err
      dtype <- (#peek FIELD_DEF, dtype) buffer
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, dsize) buffer) nullPtr (#const OCI_ATTR_DATA_SIZE) err >>= handleSqlResult err
      dsize <- (#peek FIELD_DEF, dsize) buffer
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, dprec) buffer) nullPtr (#const OCI_ATTR_PRECISION) err >>= handleSqlResult err
      dprec <- (#peek FIELD_DEF, dprec) buffer
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, dscale) buffer) nullPtr (#const OCI_ATTR_SCALE) err >>= handleSqlResult err
      dscale <- (#peek FIELD_DEF, dscale) buffer
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, isNull) buffer) nullPtr (#const OCI_ATTR_IS_NULL) err >>= handleSqlResult err
      isNull <- (#peek FIELD_DEF, isNull) buffer
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, colName) buffer) ((#ptr FIELD_DEF, colNameLen) buffer) (#const OCI_ATTR_NAME) err >>= handleSqlResult err
      pColName <- (#peek FIELD_DEF, colName) buffer
      colNameLen <- (#peek FIELD_DEF, colNameLen) buffer
      colName <- peekCStringLen (pColName, fromIntegral (colNameLen :: (#type ub4)))
      ociDescriptorFree par (#const OCI_DTYPE_PARAM)
      return (colName,mkSqlType dtype dsize dprec dscale,toBool (fromIntegral (isNull :: (#type ub1))))
            
    beginTransaction myEnvironment svcCtx err =
      ociTransStart svcCtx err 0 (#const OCI_TRANS_READWRITE) >>= handleSqlResult err
    
    commitTransaction myEnvironment svcCtx err =
      ociTransCommit svcCtx err (#const OCI_DEFAULT) >>= handleSqlResult err
    
    rollbackTransaction myEnvironment svcCtx err = do
      ociTransRollback svcCtx err (#const OCI_DEFAULT) >>= handleSqlResult err

    closeStatement stmt buffer err = do
      ociHandleFree stmt typeIdent >>= handleSqlResult err
      free buffer

    setPrefetch stmt err nrows =
      alloca $ \pNRows -> do
        poke pNRows nrows
        res <- ociAttrSet stmt typeIdent pNRows 0 (#const OCI_ATTR_PREFETCH_ROWS) err
        handleSqlResult err res

    fetch stmt err = do
      res <- ociStmtFetch2 stmt err 1 (#const OCI_FETCH_NEXT) 0 (#const OCI_DEFAULT)
      handleSqlResult err res
      return (res /= OCIResult (#const OCI_NO_DATA))

    getColValue :: Ptr () -> Int -> FieldDef -> (FieldDef -> CString -> Int -> IO a) -> IO a
    getColValue buffer colNumber fieldDef f = do
      offset <- peek (castPtr buffer `advancePtr` colNumber) :: IO CInt
      let nullIndPtr = castPtr buffer `plusPtr` (fromIntegral offset)
          valuePtr   = castPtr buffer `plusPtr` (fromIntegral offset + sizeOf (undefined :: CShort))
      nullInd  <- peek nullIndPtr :: IO CShort
      if nullInd /= 0
        then f fieldDef nullPtr 0
        else do valueLen <- strlen valuePtr
                f fieldDef valuePtr (fromIntegral valueLen)


{- |
For older Oracle databases.
-}
tablesFromCoreDB ::
     OCIHandle b
  -> OCISvcCtx
  -> OCIError
  -> IO [String]
tablesFromCoreDB env svcCtx err =
  withCStringLen "COREDB_SYSTEM" $ \(cstr,clen) ->
  withHandle env err $ \descr -> do
  ociDescribeAny svcCtx err (castPtr cstr) (fromIntegral clen) (#const OCI_OTYPE_NAME) (#const OCI_DEFAULT) (#const OCI_PTYPE_SCHEMA) descr >>= handleSqlResult err
  withAttr descr err (#const OCI_ATTR_PARAM) $ \param ->
    withAttr (param::OCIParam) err (#const OCI_ATTR_LIST_OBJECTS) $ \coll ->
        allocaBytes (#const (sizeof(FIELD_DEF))) (getTableNames coll 1 err)



getTableNames ::
  OCIParam -> CInt -> Ptr a -> OCIError -> IO [String]
getTableNames coll index buffer err = do
  res <- ociParamGet coll typeIdent err ((#ptr FIELD_DEF, par) buffer) index
  par <- (#peek FIELD_DEF, par) buffer
  if res == OCIResult (#const OCI_SUCCESS)
    then do
      ociAttrGet (par::OCIParam) typeIdent ((#ptr FIELD_DEF, colName) buffer) ((#ptr FIELD_DEF, colNameLen) buffer) (#const OCI_ATTR_OBJ_NAME) err >>= handleSqlResult err
      pName <- (#peek FIELD_DEF, colName) buffer
      nameLen <- (#peek FIELD_DEF, colNameLen) buffer
      name <- peekCStringLen (pName, fromIntegral (nameLen :: (#type ub4)))
      ociAttrGet par typeIdent ((#ptr FIELD_DEF, dtype) buffer) nullPtr (#const OCI_ATTR_PTYPE) err >>= handleSqlResult err
      ptype <- (#peek FIELD_DEF, dtype) buffer
      ociDescriptorFree par (#const OCI_DTYPE_PARAM)
      names <- getTableNames coll (index+1) buffer err
      return $! (if ptype == ((#const OCI_PTYPE_TABLE) :: (#type ub1))
                   then name:names
                   else      names)
    else return []



{- |
For newer Oracle databases, e.g. 10g.
-}
tablesFromUserTables :: IO Connection -> IO [String]
tablesFromUserTables connector =
  bracket connector HSQL.disconnect retrieveTables

retrieveTables :: HSQL.Connection -> IO [String]
retrieveTables c =
  HSQL.query c "select table_name from user_tables" >>=
    HSQL.collectRows (flip HSQL.getFieldValue "TABLE_NAME")




-- auxiliary functions


withAttr :: OCIRec c =>
     OCIHandle c
  -> OCIError
  -> CInt
  -> (OCIHandle b -> IO a)
  -> IO a
withAttr param err attrType =
  bracketSQL err
    (\pAttr -> ociAttrGet param typeIdent pAttr nullPtr attrType err)
    (\attr -> ociDescriptorFree attr (#const OCI_DTYPE_PARAM))

withHandle :: OCIRec b =>
     OCIHandle c
  -> OCIError
  -> (OCIHandle b -> IO a)
  -> IO a
withHandle env err =
  bracketSQL err
    (\pHandle -> ociHandleAlloc env pHandle typeIdent 0 nullPtr)
    (\handle -> ociHandleFree handle typeIdent)

bracketSQL :: Storable handle =>
     OCIError
  -> (Ptr handle -> IO OCIResult)
  -> (handle -> IO OCIResult)
  -> (handle -> IO a)
  -> IO a
bracketSQL err open close action =
  bracket
    (create err open)
    (\h -> close h >>= handleSqlResult err)
    action

create :: Storable handle =>
  OCIError -> (Ptr handle -> IO OCIResult) -> IO handle
create err open =
  alloca $ \p -> open p >>= handleSqlResult err >> peek p
